unit FileInfo;

{
1997 Mar 01  Updated to know about NTFS disk and file compression
1997 May 23  Change Directory to Folder....
             Replace most GetXXX functions with read-only properties
1997 Oct 04  Add maximum file name length determination
1997 Oct 18  Ensure FindClose is called even if a stop is requested
             Separate out the AddFile method for TFolderList
1997 Oct 26  Collect data for count of files
1997 Dec 05  Use double to handle large disks better
              - assume there are less than 2GB (count) of files or folders
}

interface

uses
  Classes;

type
  TScanCallback = procedure (const status: string); stdcall;

  // forward class declaration
  TFolderData = class;

  TFolderList = class (TStringList)
  private
    Ftotal_bytes: double;                 // count of bytes here and below
    Ftotal_allocated: double;             // count of allocated bytes
    Ftotal_slack: double;                 // count of slack bytes
    Ftotal_files: integer;                // count of files here and below
    Ftotal_folders: integer;              // count of folders here and below
    Ffolder_name: string;                 // full path specification
    Fparent_folder: TFolderList;          // pointer to previous folder, or nil
    Fmax_path_string: string;
    Fmax_path_length: integer;            // max full path spec length
  public
    constructor Create (const Parent: TFolderList;  const Name: string);
    procedure SetFolderName (const Name: string);
    procedure scan (var stop_requested: boolean;
                    const cluster_size: integer;
                    const show_compressed: boolean;
                    Callback: TScanCallback);
    procedure AddFile (const Name: string;
                       nominal_size: double;  true_size: double);
    property TotalBytes: double read Ftotal_bytes;
    property TotalAllocated: double read Ftotal_allocated;
    property TotalSlack: double read Ftotal_slack;
    property TotalFolders: integer read Ftotal_folders;
    property TotalFiles: integer read Ftotal_files;
    property ParentFolderList: TFolderList read Fparent_folder;
    property FolderName: string read Ffolder_name;
    property MaxPathLength: integer read Fmax_path_length;
    property MaxPathString: string read Fmax_path_string;
  end;

  TFolderData = class
  private
    Fbytes: double;
    Fallocated: double;
    Fslack: double;
    Ffiles: integer;              // number of files in this folder and below
    Fowner_folder: TFolderList;   // pointer to current folder list, or nil
    Fsub_folder: TFolderList;     // pointer to sub-folder list, or nil
    function GetParentFolderList: TFolderList;
  public
    constructor Create (const Size: double;
                        const Allocated: double;
                        const NumberOfFiles: integer;
                        const OwnerFolderList: TFolderList;
                        const SubFolderList: TFolderList);
    destructor Destroy;  override;
    property ParentFolderList: TFolderList read GetParentFolderList;
    property Bytes: double read Fbytes;
    property Allocated: double read Fallocated;
    property Slack: double read Fslack;
    property Files: integer read Ffiles;
    property SubFolderList: TFolderList read Fsub_folder;
  end;


implementation

uses
  SysUtils, Forms, Windows;

// methods for TFolderData
constructor TFolderData.Create (const Size: double;
                                   const Allocated: double;
                                   const NumberOfFiles: integer;
                                   const OwnerFolderList: TFolderList;
                                   const SubFolderList: TFolderList);
begin
  inherited Create;
  Fowner_folder := OwnerFolderList;
  Fsub_folder := SubFolderList;
  Fbytes := Size;
  Fallocated := Allocated;
  Fslack := Allocated - Size;
  Ffiles := NumberOfFiles;
end;

destructor TFolderData.Destroy;
// must dispose of sub-folders as well as main entry
begin
  if Fsub_folder <> nil then
    begin
    Fsub_folder.Destroy;
    Fsub_folder := nil;
    end;
  Inherited Destroy;
end;

function TFolderData.GetParentFolderList: TFolderList;
begin
  Result := Fowner_folder.ParentFolderList;
end;


// methods for TFolderList
constructor TFolderList.Create (const Parent: TFolderList;  const Name: string);
// as standard string list, but allow duplicates, stores name and back-link
begin
  inherited Create;
  Sorted := False;
  Ftotal_folders := 0;
  Ftotal_bytes := 0.0;
  Ftotal_allocated := 0.0;
  Ftotal_slack := 0.0;
  Ftotal_files := 0;
  Fparent_folder := Parent;
  Ffolder_name := Name;
  Fmax_path_string := '';
  Fmax_path_length := 0;
end;

procedure TFolderList.SetFolderName (const Name: string);
begin
  Clear;
  Ffolder_name := Name;
end;

procedure TFolderList.scan (var stop_requested: boolean;
                               const cluster_size: integer;
                               const show_compressed: boolean;
                               Callback: TScanCallback);

  function allocated_bytes (file_size: double;  cluster_size: integer): double;
  // this code should work for files up to 2GB worth of clusters....
  var
     fill: integer;
     clusters: double;
  begin
    Result := file_size;
    if cluster_size <> 1 then
      begin
      fill := cluster_size - 1;     // one less than the capacity of a cluster
      file_size := file_size + fill;  // push to the next cluster up
      clusters := trunc (file_size / cluster_size);
      Result := clusters * cluster_size;
      end;
  end;

var
   s: TSearchRec;
   sub_folder_name: string;
   sub_folder: TFolderList;
   next_entry: TFolderData;
   bytes_below: double;
   allocated_below: double;
   slack_below: double;
   folders_below: integer;
   files_below: integer;
   true_size: double;
   nominal_size: double;
   max_path: string;
   path_length: integer;
begin
  Ftotal_bytes := 0.0;
  Ftotal_allocated := 0.0;
  Ftotal_slack := 0.0;
  Ftotal_folders := 0;
  Ftotal_files := 0;
  max_path := '';
  Fmax_path_length := 0;
  Callback ('Scanning ' + Ffolder_name + '...');
  Application.ProcessMessages;
  s.FindHandle := INVALID_HANDLE_VALUE;
  try
    if FindFirst (Ffolder_name + '*.*', faAnyFile, s) <> 0
    then s.FindHandle := INVALID_HANDLE_VALUE
    else
    repeat
      with s do
        begin
        if (Attr and faDirectory) <> 0
        then
          begin
          // file is sub-folder - but ignore the parent and local backlink
          if (Name = '.') or (Name = '..')
          then
          else
            begin
            Inc (Ftotal_folders);
            // form new full sub-folder name, with the trailing '\'
            sub_folder_name := Ffolder_name + Name + '\';
            // allocate a new folder list, insert a record
            // emulating the parent folder
            sub_folder := TFolderList.Create (Self, sub_folder_name);
            sub_folder.AddObject ('..', TFolderData.Create (0, 0, 0, sub_folder, nil));
            // make the inserted name start with a '\'
            sub_folder_name := '\' + Name;
            // allocate a new entry for the current scan, find out
            // how many bytes and files are in the sub-tree
            next_entry := TFolderData.Create (Size, Size, 0, Self, sub_folder);
            with sub_folder do
              begin
              scan (stop_requested, cluster_size, show_compressed, Callback);
              bytes_below := Ftotal_bytes;
              allocated_below := Ftotal_allocated;
              slack_below := Ftotal_slack;
              folders_below := Ftotal_folders;
              files_below := Ftotal_files;
              // set the size found in this sub-tree and insert the
              // new entry into the current list
              next_entry.Fbytes := bytes_below;
              next_entry.Fallocated := allocated_below;
              next_entry.Fslack := slack_below;
              next_entry.Ffiles := files_below;
              path_length := MaxPathLength;     // find out the maximum size
              max_path := MaxPathString;        // and where it occured
              end;
            AddObject (sub_folder_name, next_entry);
            Ftotal_bytes := Ftotal_bytes + bytes_below;    // bump total bytes found at this
            Ftotal_allocated := Ftotal_allocated + allocated_below;    // (and lower) levels
            Ftotal_slack := Ftotal_slack + slack_below;
            Inc (Ftotal_folders, folders_below);
            Inc (Ftotal_files, files_below);
            if path_length > Fmax_path_length then
              begin
              Fmax_path_length := path_length;
              Fmax_path_string := max_path;
              end;
            end;
          end
        else
          begin
          // Compute actual size on disk.  This is either the real file size,
          // quantised upwards by the cluster size or, for NTFS compressed files
          // it is the compressed size similarly quantised.  Note that the
          // GetCompressedFileSize call is not available in Windows 95.
          nominal_size := Size;
          if (show_compressed) and
             ((FindData.dwFileAttributes and FILE_ATTRIBUTE_COMPRESSED) <> 0)
          then nominal_size := GetCompressedFileSize (
                              PChar (Ffolder_name + FindData.cFileName), nil);
          if nominal_size = -1 then nominal_size := Size;

          true_size := allocated_bytes (nominal_size, cluster_size);
          // insert into the list
          AddFile (Name, nominal_size, true_size);
          max_path := Ffolder_name + FindData.cFileName;
          path_length := Length (max_path);
          if path_length > Fmax_path_length then
            begin
            Fmax_path_length := path_length;
            Fmax_path_string := max_path;
            end;
          end;
        end;
    until (FindNext (s) <> 0) or (stop_requested);
  finally
    // after adding the NT compression code,
    // needed to add SysUtils to the line below!
    SysUtils.FindClose (s);
  end;
end;

procedure TFolderList.AddFile (const Name: string;
                       nominal_size: double;  true_size: double);
begin
  AddObject (Name, TFolderData.Create (nominal_size, true_size, 1, Self, nil));
  Ftotal_bytes := Ftotal_bytes + nominal_size;
  Ftotal_allocated := Ftotal_allocated + true_size;
  Ftotal_slack := Ftotal_slack + true_size - nominal_size;
  Inc (Ftotal_files);
end;


end.

